/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>

#include <fstream>

#include <voms_api.h>
#include "data.h"

/*
extern bool contact(const std::string &hostname, int port, 
		    const std::string &contact,
		    const std::string &command, std::string &buffer, 
		    std::string &subject, std::string &ca, verror_type &error);
extern bool retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how, 
		     std::string &buffer, std::string &vo, std::string &file, 
		     std::string &subject, std::string &ca, verror_type &error);
extern bool verify(std::string message, vomsdata &voms, verror_type &error, 
		   std::string vdir, std::string cdir, std::string subject, 
		   std::string ca);
*/
#include "api_util.h"

vomsdata::
vomsdata::vomsdata(std::string voms_dir, std::string cert_dir) :  ca_cert_dir(cert_dir),
								  voms_cert_dir(voms_dir),
								  duration(0),
								  ordering(","),
								  error(VERR_NONE),
								  workvo(""),
								  extra_data("")
{
  if (voms_cert_dir.empty()) {
    char *v;
    if ( (v = getenv("X509_VOMS_DIR")))
      voms_cert_dir = std::string(v);
    else 
      voms_cert_dir = "/etc/grid-security/vomsdir";
  }

  if (ca_cert_dir.empty()) {
    char *c;
    if ((c = getenv("X509_CERT_DIR")))
      ca_cert_dir = std::string(c);
    else
      ca_cert_dir = "/etc/grid-security/certificates";
  }

  DIR *vdir, *cdir;
  vdir = opendir(voms_cert_dir.c_str());
  cdir = opendir(ca_cert_dir.c_str());

  if (!vdir || !cdir)
    error = VERR_DIR;
    
  if (cdir)
    (void)closedir(cdir);
  if (vdir)
    (void)closedir(vdir);

  duration = 0;
}

void vomsdata::SetLifetime(int lifetime)
{
  duration = lifetime;
}

void vomsdata::ResetOrder(void)
{
  ordering=",";
}

void vomsdata::Order(std::string att)
{
  ordering += att + ",";
}

bool vomsdata::Contact(std::string hostname, int port, std::string servsubject, std::string command)
{
  std::string buffer;
  std::string subject, ca;
  std::string lifetime;

  if (contact(hostname, port, servsubject, stringify(duration, lifetime) + ordering + command, buffer, subject, ca, error)) {
    while (buffer[0] == '1') {
      std::string::size_type t = buffer.find('\1');
      if (t != std::string::npos)
	buffer = buffer.substr(t+1);
      else {
	error = VERR_FORMAT;
	return false;
      }
    }

    while (buffer[0] == '0') {
      std::string::size_type t = buffer.find('\1');
      if (t != std::string::npos) {
	error = VERR_ORDER;
	buffer = buffer.substr(t+1);
      }
      else {
	error = VERR_FORMAT;
	return false;
      }
    }

    while (buffer[0] == '2') {
      std::string::size_type t = buffer.find('\1');
      if (t != std::string::npos)
	error = VERR_NODATA;
      else
	error = VERR_FORMAT;
      return false;
    }

    return verify(buffer, *this, error, voms_cert_dir, ca_cert_dir, subject, ca);
  }

  return false;
}

bool vomsdata::Retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how)
{
  std::string buffer;
  std::string vo;
  std::string file;
  std::string subject;
  std::string ca;

  if (retrieve(cert, chain, how, buffer, vo, file, subject, ca, error)) {
    workvo = vo;
    extra_data = file;
    return verify(buffer, *this, error, voms_cert_dir, ca_cert_dir, subject, ca);
  }
  return false;
}

bool vomsdata::Import(std::string buffer)
{
  std::string dummy="";

  return verify(buffer, *this, error, voms_cert_dir, ca_cert_dir, dummy, dummy);
}

bool vomsdata::Export(std::string &buffer)
{
  std::string result;
  std::string temp;

  for (std::vector<voms>::iterator v=data.begin(); v != data.end(); v++) {
    int len;
    std::string d;

    result += std::string("SIGLEN:") + stringify(v->siglen, temp) +
      "\nSIGNATURE:" + v->signature + "USER:" + v->user + "\nUCA:" + 
      v->userca + "\nSERVER:" + v->server + "\nSCA:" + v->serverca + 
      "\nVO:" + v->voname + "\nURI:" + v->uri + "\nTIME1:" + v->date1 + 
      "\nTIME2:" + v->date2 + "\nDATALEN:";

    switch (v->type) {
    case TYPE_NODATA:
      d = std::string("NO DATA\n");
      break;
    case TYPE_CUSTOM:
      d = v->custom;
      break;
    case TYPE_STD:
      for (std::vector<struct data> ::iterator dt = v->std.begin(); dt != v->std.end(); dt++)
	d += std::string("GROUP: ") + dt->group + "\nROLE: " + dt->role + 
	  "\nCAP: " + dt->cap + "\n";
      break;
    default:
      error = VERR_TYPE;
      return false;
      break;
    }
    len = d.size();

    result += stringify(len, temp) + "\n" + d;
  }

  buffer = result;
  return true;
}

bool vomsdata::DefaultData(voms &d)
{
  if (data.empty()) {
    error = VERR_NOEXT;
    return false;
  }

  d = data.front();
}

bool vomsdata::loadfile(std::string filename, uid_t uid, gid_t gid)
{
  struct stat stats;

  int error = 0;

  struct vomsdata data;

  if (filename.empty()) {
    error = VERR_DIR;
    return false;
  }

  if (stat(filename.c_str(), &stats) == -1 || stats.st_uid != uid ||
      stats.st_gid != gid ||
      !(stats.st_mode == (S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) ||
       stats.st_mode == (S_IFDIR |
			 S_IRUSR | S_IWUSR | S_IXUSR |
			 S_IRGRP | S_IXGRP |
			 S_IROTH | S_IXOTH))) {
    error = VERR_DIR;
    return false;
  }

  if (stats.st_mode & S_IFREG)
    return loadfile0(filename, uid, gid);
  else {
    DIR *dp = opendir(filename.c_str());
    struct dirent *de;
    bool cumulative = false;

    if (dp) {
      while (de = readdir(dp)) {
	char *name = de->d_name;
	if (name && (strcmp(name, ".") != 0) && (strcmp(name, "..") != 0))
	  cumulative |= loadfile(filename + "/" + name, uid, gid);
      }
      return cumulative;
    }
  }
  return false;
}

static bool
tokenize(std::string str, unsigned int &start, std::string &value)
{
  if (start != std::string::npos) {
    unsigned int begin = str.find('"',start);
    if (begin != std::string::npos) {
      unsigned int end = str.find('"',begin+1);
      if (end != std::string::npos) {
	value = str.substr(begin+1, end-begin-1);
	start = end+1;
	if (start >= str.size())
	  start = std::string::npos;
	return true;
      }
    }
  }
  return false;
}

bool vomsdata::loadfile0(std::string filename, uid_t uid, gid_t gid)
{
  struct contactdata data;

  if (filename.empty()) {
    error = VERR_DIR;
    return false;
  }

  /* Opens the file */
  std::ifstream f(filename.c_str());

  if (!f) {
    error = VERR_DIR;
    return false;
  }

  /* Load the file */
  int linenum = 1;
  bool ok = true;

  while (ok && f) {
    std::string line;

    if (getline(f,line)) {
      unsigned int start = 0;
      std::string port;

      ok &= tokenize(line, start, data.nick);
      ok &= tokenize(line, start, data.host);
      ok &= tokenize(line, start, port);
      ok &= tokenize(line, start, data.contact);
      ok &= tokenize(line, start, data.vo);

      data.port = atoi(port.c_str());
      servers.push_back(data);
    }
    else {
      error = VERR_FORMAT;
      return false;
    }
    linenum++;
  }
  return true;
}

bool vomsdata::LoadSystemContacts(std::string dir)
{
  if (dir.empty())
    dir = "/opt/edg/etc/vomses";

  return loadfile(dir, 0, 0);
}

bool vomsdata::LoadUserContacts(std::string dir)
{
  if (dir.empty()) {
    char *name = getenv("VOMS_USERCONF");
    if (name)
      dir = std::string(name);
    else {
      char *home = getenv("HOME");
      if (home)
	dir = std::string(home) + "/.edg/vomses";
      else
	dir = "~/.edg/vomses";
    }
  }

  return loadfile(dir, getuid(), getgid());
}

std::vector<contactdata>
vomsdata::FindByAlias(std::string nick)
{
  std::vector<contactdata>::iterator beg = servers.begin(), end = servers.end();
  std::vector<contactdata> results;

  while (beg != end) {
    if (beg->nick == nick)
      results.push_back(*beg);
    beg++;
  }

  return results;
}

std::vector<contactdata> vomsdata::FindByVO(std::string vo)
{
  std::vector<contactdata>::iterator beg = servers.begin(), end = servers.end();
  std::vector<contactdata> results;

  while (beg != end) {
    if (beg->vo == vo)
      results.push_back(*beg);
    beg++;
  }

  return results;
}


int getMajorVersionNumber(void) {return 1;}
int getMinorVersionNumber(void) {return 4;}
int getPatchVersionNumber(void) {return 0;}
